from __future__ import annotations

from typing import TYPE_CHECKING
from qtpy import QtWidgets as QtW, QtCore, QtGui
from himena import WidgetDataModel, StandardType
from himena.plugins import validate_protocol
from himena_stats.distributions._utils import draw_pdf_or_pmf, infer_edges

if TYPE_CHECKING:
    from scipy import stats


class QDistGraphics(QtW.QGraphicsView):
    """Graphics view for displaying a distribution."""

    def __init__(self):
        scene = QtW.QGraphicsScene()
        super().__init__(scene)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self._color = QtGui.QColor(128, 128, 128)
        self.setTransform(QtGui.QTransform().scale(1, -1))  # upside down

    def set_dist(self, dist: stats.rv_frozen):
        scene = self.scene()
        scene.clear()
        xlow, xhigh = infer_edges(dist)
        x, y = draw_pdf_or_pmf(dist, xlow, xhigh)
        polygon = QtGui.QPolygonF([QtCore.QPointF(x[i], y[i]) for i in range(len(x))])
        scene.addPolygon(polygon, QtGui.QPen(self._color, 0), QtGui.QBrush(self._color))
        self.fit_item()

    def resizeEvent(self, event):
        self.fit_item()
        super().resizeEvent(event)

    def fit_item(self):
        self.fitInView(self.scene().itemsBoundingRect())


class QDistParameters(QtW.QPlainTextEdit):
    def __init__(self):
        super().__init__()
        self.setWordWrapMode(QtGui.QTextOption.WrapMode.NoWrap)
        self.setReadOnly(True)

    def set_dist(self, dist: stats.rv_frozen):
        dist_name = dist.dist.name
        params = [f"{k} = {v}" for k, v in dist.kwds.items()]
        newline = "\n".join(params)
        self.setPlainText(f"{dist_name}\n\n{newline}")


class QDistributionView(QtW.QSplitter):
    def __init__(self):
        super().__init__(QtCore.Qt.Orientation.Vertical)
        self._img_view = QDistGraphics()
        self._param_view = QDistParameters()
        self._dist: stats.rv_frozen | None = None
        self.addWidget(self._img_view)
        self.addWidget(self._param_view)
        self.setSizes([100, 60])

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        dist: stats.rv_frozen = model.value
        if dist.args:
            raise NotImplementedError
        self._img_view.set_dist(dist)
        self._param_view.set_dist(dist)
        self._dist = dist

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        if self._dist is None:
            raise ValueError("No distribution is set.")
        return WidgetDataModel(
            value=self._dist,
            type=self.model_type(),
        )

    @validate_protocol
    def model_type(self) -> str:
        return StandardType.DISTRIBUTION

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 220, 200



from __future__ import annotations

from himena import WidgetDataModel
from himena.utils.collections import OrderedSet
from himena.utils.table_selection import (
    model_to_col_val_arrays,
    model_to_vals_arrays,
    NamedArray,
)
import numpy as np


def dropna(arr: NamedArray) -> np.ndarray:
    mask = np.isnan(arr.array)
    return arr.array[~mask]


def pvalue_to_asterisks(pval: float) -> str:
    if pval > 0.05:
        return "n.s."
    elif pval > 0.01:
        return "*"
    elif pval > 0.001:
        return "**"
    elif pval > 0.0001:
        return "***"
    else:
        return "****"


def values_groups_to_arrays(
    model: WidgetDataModel,
    values: list[tuple[tuple[int, int], tuple[int, int]]],
    groups,
) -> list[NamedArray]:
    if groups is None:
        arrs = model_to_vals_arrays(
            model,
            values,
            same_size=False,
        )
    else:
        if len(values) != 1:
            raise ValueError("If groups are given, values must be a single range.")
        col, val = model_to_col_val_arrays(model, groups, values[0])
        unique_values = OrderedSet(col.array)
        arrs = [
            NamedArray(str(uval), val.array[col.array == uval])
            for uval in unique_values
        ]
    return arrs


def values_groups_to_xy(
    model: WidgetDataModel,
    values: list[tuple[tuple[int, int], tuple[int, int]] | None],
    groups,
) -> tuple[NamedArray, NamedArray]:
    values = [val for val in values if val is not None]
    arrs = values_groups_to_arrays(model, values, groups)
    if len(arrs) != 2:
        raise ValueError(f"Expected two groups, found {len(arrs)}")
    return arrs[0], arrs[1]



from himena import Parametric, StandardType, WidgetDataModel
from himena.widgets import SubWindow
from himena.qt.magicgui import SelectionEdit
from himena.plugins import register_function, configure_gui
from himena.utils.table_selection import model_to_vals_arrays, range_getter
import numpy as np
from himena_stats.consts import MENUS_DIST, TABLE_LIKE
from himena_stats._lazy_import import stats


@register_function(
    menus=MENUS_DIST,
    title="Normal Distribution ...",
    command_id="himena-stats:dist-construct:continuous:norm",
)
def dist_norm() -> Parametric:
    """Construct normal distribution."""

    @configure_gui
    def construct_norm(mu: float = 0.0, sigma: float = 1.0):
        return WidgetDataModel(
            value=stats.norm(loc=mu, scale=sigma),
            type=StandardType.DISTRIBUTION,
            title="Normal",
        )

    return construct_norm


@register_function(
    menus=MENUS_DIST,
    title="Uniform Distribution ...",
    command_id="himena-stats:dist-construct:continuous:uniform",
)
def dist_uniform() -> Parametric:
    """Construct uniform distribution."""

    @configure_gui
    def construct_uniform(a: float = 0.0, b: float = 1.0):
        return WidgetDataModel(
            value=stats.uniform(loc=a, scale=b - a),
            type=StandardType.DISTRIBUTION,
            title="Uniform",
        )

    return construct_uniform


@register_function(
    menus=MENUS_DIST,
    title="Exponential Distribution ...",
    command_id="himena-stats:dist-construct:continuous:expon",
)
def dist_expon() -> Parametric:
    """Construct exponential distribution."""

    @configure_gui
    def construct_expon(scale: float = 1.0):
        return WidgetDataModel(
            value=stats.expon(scale=scale),
            type=StandardType.DISTRIBUTION,
            title="Exponential",
        )

    return construct_expon


@register_function(
    menus=MENUS_DIST,
    title="Gamma Distribution ...",
    command_id="himena-stats:dist-construct:continuous:gamma",
)
def dist_gamma() -> Parametric:
    """Construct Gamma distribution."""

    @configure_gui
    def construct_gamma(a: float = 1.0, scale: float = 1.0):
        return WidgetDataModel(
            value=stats.gamma(a=a, scale=scale),
            type=StandardType.DISTRIBUTION,
            title="Gamma",
        )

    return construct_gamma


@register_function(
    menus=MENUS_DIST,
    title="Beta Distribution ...",
    command_id="himena-stats:dist-construct:continuous:beta",
)
def dist_beta() -> Parametric:
    """Construct Beta distribution."""

    @configure_gui
    def construct_beta(a: float = 2.0, b: float = 2.0):
        return WidgetDataModel(
            value=stats.beta(a=a, b=b),
            type=StandardType.DISTRIBUTION,
            title="Beta",
        )

    return construct_beta


@register_function(
    menus=MENUS_DIST,
    title="Cauchy Distribution ...",
    command_id="himena-stats:dist-construct:continuous:cauchy",
)
def dist_cauchy() -> Parametric:
    """Construct Cauchy distribution."""

    @configure_gui
    def construct_cauchy(loc: float = 0.0, scale: float = 1.0):
        return WidgetDataModel(
            value=stats.cauchy(loc=loc, scale=scale),
            type=StandardType.DISTRIBUTION,
            title="Cauchy",
        )

    return construct_cauchy


@register_function(
    menus=MENUS_DIST,
    title="T Distribution ...",
    command_id="himena-stats:dist-construct:continuous:t",
)
def dist_t() -> Parametric:
    """Construct t distribution."""

    @configure_gui
    def construct_t(df: float = 1.0):
        return WidgetDataModel(
            value=stats.t(df=df),
            type=StandardType.DISTRIBUTION,
            title="T",
        )

    return construct_t


@register_function(
    menus=MENUS_DIST,
    title="Binomial Distribution ...",
    command_id="himena-stats:dist-construct:discrete:binom",
)
def dist_binom() -> Parametric:
    """Construct binomial distribution."""

    @configure_gui
    def construct_binom(n: int = 10, p: float = 0.5):
        return WidgetDataModel(
            value=stats.binom(n=n, p=p),
            type=StandardType.DISTRIBUTION,
            title="Binomial",
        )

    return construct_binom


@register_function(
    menus=MENUS_DIST,
    title="Poisson Distribution ...",
    command_id="himena-stats:dist-construct:discrete:poisson",
)
def dist_poisson() -> Parametric:
    """Construct Poisson distribution."""

    @configure_gui
    def construct_poisson(mu: float = 5.0):
        return WidgetDataModel(
            value=stats.poisson(mu=mu),
            type=StandardType.DISTRIBUTION,
            title="Poisson",
        )

    return construct_poisson


@register_function(
    menus=MENUS_DIST,
    types=TABLE_LIKE,
    title="Empirical Distribution ...",
    command_id="himena-stats:dist-construct:empirical",
)
def dist_empirical(win: SubWindow) -> Parametric:
    @configure_gui(
        selection={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_dist_emprifical(
        selection: tuple[tuple[int, int], tuple[int, int]],
        bins: int = 10,
    ) -> WidgetDataModel:
        model = win.to_model()
        arr = model_to_vals_arrays(
            model,
            [selection],
        )[0].array.ravel()
        hist = stats.rv_histogram(np.histogram(arr, bins=bins), density=False).freeze()
        return WidgetDataModel(
            value=hist,
            type=StandardType.DISTRIBUTION,
            title="Empirical",
        )

    return run_dist_emprifical



from himena.plugins import register_widget_class
from himena import StandardType
from himena_stats.distributions import _construct, _methods
from himena_stats.distributions._widget import QDistributionView


def _register_widgets():
    register_widget_class(StandardType.DISTRIBUTION, QDistributionView)


_register_widgets()

del _construct, _methods, _register_widgets



from typing import Literal
from himena import StandardType
from himena.plugins import configure_submenu

MENUS_TEST = ["tools/stats-test", "/model_menu/stats-test"]
MENUS_DIST = ["tools/stats-dist", "/model_menu"]
configure_submenu(MENUS_TEST, title="Statistical Tests")
configure_submenu(MENUS_DIST, title="Distributions")

TABLE_LIKE = [StandardType.TABLE, StandardType.DATAFRAME, StandardType.EXCEL]
ALTERNATIVE = Literal["two-sided", "less", "greater"]



import pytest

@pytest.fixture(scope="session", autouse=True)
def install_plugin(request: pytest.FixtureRequest):
    import himena_stats.distributions  # noqa: F401
    import himena_stats.io  # noqa: F401
    import himena_stats.test_tools  # noqa: F401



from __future__ import annotations
from typing import Literal

from himena import Parametric, StandardType, WidgetDataModel
from himena.widgets import SubWindow
from himena.plugins import configure_gui
from himena.utils.table_selection import range_getter, model_to_vals_arrays
from himena.qt.magicgui import SelectionEdit

from himena_stats._lazy_import import stats
from himena_stats.consts import ALTERNATIVE
from himena_stats.test_tools._utils import pvalue_to_asterisks


def shapiro_wilk_test(win: SubWindow) -> Parametric:
    """Run a Shapiro-Wilk test on a table-like data."""

    @configure_gui(
        sample_data={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_shapiro_wilk_test(sample_data: tuple[tuple[int, int], tuple[int, int]]):
        model = win.to_model()
        x0 = model_to_vals_arrays(model, [sample_data])[0]
        sw_result = stats.shapiro(x0.array)
        sw_result_table = [
            ["p-value", format(sw_result.pvalue, ".5g")],
            ["", pvalue_to_asterisks(sw_result.pvalue)],
            ["statistic", format(sw_result.statistic, ".5g")],
        ]
        return WidgetDataModel(
            value=sw_result_table,
            type=StandardType.TABLE,
            title=f"Shapiro-Wilk Test result of {model.title}",
        )

    return run_shapiro_wilk_test


def kolmogorov_smirnov_test(win: SubWindow) -> Parametric:
    """Run a Kolmogorov-Smirnov test on a table-like data."""

    @configure_gui(
        sample_data={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_kolmogorov_smirnov_test(
        sample_data: tuple[tuple[int, int], tuple[int, int]],
        cdf: Literal["norm", "expon", "uniform"] = "norm",
        alternative: ALTERNATIVE = "two-sided",
        method: Literal["auto", "exact", "approx", "asymp"] = "auto",
    ):
        """
        Run Kolmogorov-Smirnov test on a table-like data.

        Parameters
        ----------
        sample_data : tuple[tuple[int, int], tuple[int, int]]
            Range of the sample data.
        cdf : function, optional
            Cumulative distribution function, by default "norm"
        alternative : str, default "two-sided"
            Test alternative hypothesis. This parameteris forwarded to
            `scipy.stats.kstest`.
        method : str, optional
            The method to use to calculate the p-value. This parameteris forwarded to
            `scipy.stats.kstest`. Following is from the documentation of `stats.kstest`:

            * 'auto' : selects one of the other options.
            * 'exact' : uses the exact distribution of test statistic.
            * 'approx' : approximates the two-sided probability with twice the
                one-sided probability
            * 'asymp': uses asymptotic distribution of test statistic, by default "auto"
        """
        model = win.to_model()
        x0 = model_to_vals_arrays(model, [sample_data])[0]
        ks_result = stats.kstest(x0.array, cdf, alternative=alternative, method=method)
        ks_result_table = [
            ["p-value", format(ks_result.pvalue, ".5g")],
            ["", pvalue_to_asterisks(ks_result.pvalue)],
            ["statistic", format(ks_result.statistic, ".5g")],
        ]
        return WidgetDataModel(
            value=ks_result_table,
            type=StandardType.TABLE,
            title=f"Kolmogorov-Smirnov Test result of {model.title}",
        )

    return run_kolmogorov_smirnov_test



import pytest
from himena import MainWindow


@pytest.mark.parametrize(
    "command,params",
    [
        ("continuous:norm", {"mu": 3.2, "sigma": 4.5}),
        ("continuous:uniform", {"a": -2, "b": 3.5}),
        ("continuous:expon", {"scale": 2.1}),
        ("continuous:gamma", {"a": 3.0, "scale": 1.2}),
        ("continuous:beta", {"a": 2.0, "b": 2.0}),
        ("continuous:cauchy", {"loc": 1.0, "scale": 2.0}),
        ("continuous:t", {"df": 15}),
        ("discrete:binom", {"n": 20, "p": 0.4}),
        ("discrete:poisson", {"mu": 4.2}),
    ]
)
def test_construction_and_sampling(himena_ui: MainWindow, command: str, params: dict):
    himena_ui.exec_action(
        f"himena-stats:dist-construct:{command}", with_params=params,
    )
    win_dist = himena_ui.current_window
    himena_ui.exec_action("himena-stats:dist:stats")
    himena_ui.exec_action(
        "himena-stats:dist:sample",
        window_context=win_dist,
        with_params={"sample_size": 10, "random_state": 1234}
    )
    win_sample = himena_ui.current_window
    assert win_dist is not None
    assert win_sample is not None

    himena_ui.exec_action(
        "himna-stats:dist-convert:fit-mle",
        model_context=win_dist.to_model(),
        window_context=win_dist,
        with_params={
            "obs": win_sample.to_model(),
            "obs_range": None,
        },
    )
    himena_ui.exec_action(
        "himena-stats:dist-plot:plot",
        window_context=win_dist,
        with_params={"obs": win_sample.to_model(), "obs_range": None}
    )



from __future__ import annotations

from functools import cache
import json
from pathlib import Path

from himena import StandardType, WidgetDataModel
from himena.plugins import register_reader_plugin, register_writer_plugin
from himena_stats.io._serialize import serialize_rv, deserialize_rv
from himena_stats._lazy_import import stats


@cache
def rv_frozen_type() -> type[stats.rv_frozen]:
    """Return the stats.rv_frozen type (because rv_frozen is not public in runtime)."""
    return type(stats.uniform())


@register_reader_plugin
def read_distribution(path: Path) -> WidgetDataModel:
    with path.open() as f:
        js = json.load(f)
    rv = deserialize_rv(js)
    return WidgetDataModel(
        value=rv,
        type=StandardType.DISTRIBUTION,
        title=path.name,
    )


@read_distribution.define_matcher
def _(path: Path):
    if path.suffixes == [".dist", ".json"]:
        return StandardType.DISTRIBUTION
    return None


@register_writer_plugin
def write_distribution(model: WidgetDataModel, path: Path):
    js = serialize_rv(model.value)
    with open(path, mode="w") as f:
        json.dump(js, f)


@write_distribution.define_matcher
def _(model: WidgetDataModel, path: Path):
    return isinstance(model.value, rv_frozen_type())



from __future__ import annotations

from typing import Any
from himena_stats._lazy_import import stats


def serialize_rv(rv: stats.rv_frozen) -> dict[str, Any]:
    return {"distribution": rv.dist.name, "parameters": rv.kwds.copy()}


def deserialize_rv(rv_dict: dict[str, Any]) -> stats.rv_frozen:
    dist_name = rv_dict["distribution"]
    params = rv_dict["parameters"]
    return getattr(stats, dist_name)(**params)



from typing import TYPE_CHECKING


class LazyScipyStats:
    def __getattr__(self, key: str):
        from scipy import stats

        return getattr(stats, key)


class LazyScikitPosthocs:
    def __getattr__(self, key: str):
        import scikit_posthocs

        return getattr(scikit_posthocs, key)


if TYPE_CHECKING:
    from scipy import stats
    import scikit_posthocs
else:
    stats = LazyScipyStats()
    scikit_posthocs = LazyScikitPosthocs()

__all__ = ["stats", "scikit_posthocs"]



from __future__ import annotations

import numpy as np
from himena import Parametric, StandardType, WidgetDataModel
from himena.widgets import SubWindow
from himena.plugins import register_function, configure_gui
from himena.utils.table_selection import (
    range_getter,
)
from himena.qt.magicgui import SelectionEdit

from himena_stats._lazy_import import stats, scikit_posthocs
from himena_stats.consts import MENUS_TEST, TABLE_LIKE
from himena_stats.test_tools._utils import (
    pvalue_to_asterisks,
    values_groups_to_arrays,
    dropna,
)


@register_function(
    menus=MENUS_TEST,
    title="Steel-Dwass test ...",
    types=TABLE_LIKE,
    command_id="himena-stats:test-multi:steel-dwass",
)
def steel_dwass_test(win: SubWindow) -> Parametric:
    """Run a Steel-Dwass test on a table-like data."""
    selection_opt = {"widget_type": SelectionEdit, "getter": range_getter(win)}

    @configure_gui(
        values={
            "widget_type": "ListEdit",
            "options": selection_opt,
            "value": [None],
            "layout": "vertical",
        },
        groups=selection_opt,
    )
    def run_steel_dwass_test(values: list, groups):
        model = win.to_model()
        arrs = values_groups_to_arrays(model, values, groups)
        result = scikit_posthocs.posthoc_dscf([dropna(a) for a in arrs])
        pvalues = result.to_numpy()
        return WidgetDataModel(
            value=_pval_matrix(pvalues, columns=[a.name for a in arrs]),
            type=StandardType.TABLE,
            title=f"Steel-Dwass test result of {model.title}",
        )

    return run_steel_dwass_test


@register_function(
    menus=MENUS_TEST,
    title="Tukey's HSD test ...",
    types=TABLE_LIKE,
    command_id="himena-stats:test-multi:tukey-hsd",
)
def tukey_hsd_test(win: SubWindow) -> Parametric:
    """Run a Tukey's HSD test on a table-like data."""
    selection_opt = {"widget_type": SelectionEdit, "getter": range_getter(win)}

    @configure_gui(
        values={
            "widget_type": "ListEdit",
            "options": selection_opt,
            "value": [None],
            "layout": "vertical",
        },
        groups=selection_opt,
    )
    def run_tukey_hsd_test(values: list, groups):
        model = win.to_model()
        arrs = values_groups_to_arrays(model, values, groups)
        result = stats.tukey_hsd(*[dropna(a) for a in arrs])
        return WidgetDataModel(
            value=_pval_matrix(result.pvalue, columns=[a.name for a in arrs]),
            type=StandardType.TABLE,
            title=f"Tukey HSD test result of {model.title}",
        )

    return run_tukey_hsd_test


@register_function(
    menus=MENUS_TEST,
    title="ANOVA ...",
    types=[StandardType.TABLE, StandardType.DATAFRAME, StandardType.EXCEL],
    command_id="himena-stats:test:anova",
)
def anova(win: SubWindow) -> Parametric:
    """Run an ANOVA on a table-like data."""
    selection_opt = {"widget_type": SelectionEdit, "getter": range_getter(win)}

    @configure_gui(
        values={
            "widget_type": "ListEdit",
            "options": selection_opt,
            "value": [None],
            "layout": "vertical",
        },
        groups=selection_opt,
    )
    def run_anova(values: list, groups):
        model = win.to_model()
        arrs = values_groups_to_arrays(model, values, groups)
        f_result = stats.f_oneway(*[dropna(a) for a in arrs])
        return WidgetDataModel(
            value=_pval_matrix(f_result.pvalue, columns=[a.name for a in arrs]),
            type=StandardType.TABLE,
            title=f"ANOVA result of {model.title}",
        )

    return run_anova


@register_function(
    menus=MENUS_TEST,
    title="Dunnett's test ...",
    types=TABLE_LIKE,
    command_id="himena-stats:test-multi:dunnett",
)
def dunnett_test(win: SubWindow) -> Parametric:
    """Run a Dunnett's test on a table-like data."""
    selection_opt = {"widget_type": SelectionEdit, "getter": range_getter(win)}

    @configure_gui(
        values={
            "widget_type": "ListEdit",
            "options": selection_opt,
            "value": [None],
            "layout": "vertical",
        },
        groups=selection_opt,
    )
    def run_dunnett_test(values: list, groups, control: str):
        model = win.to_model()
        arrs = values_groups_to_arrays(model, values, groups)
        if control == "":
            idx = 0
        else:
            for idx, each in enumerate(arrs):
                if each.name == control:
                    break
            else:
                raise ValueError(f"No group named {control!r}")
        treatments = [dropna(a) for a in arrs]
        control_arr = treatments.pop(idx)
        columns = [a.name for a in arrs]
        del columns[idx]
        result = scikit_posthocs.posthoc_dunnett(control_arr, treatments)
        pvalues = result.to_numpy()
        return WidgetDataModel(
            value=_pval_matrix(pvalues, columns=columns),
            type=StandardType.TABLE,
            title=f"Dunnett's test result of {model.title}",
        )

    return run_dunnett_test


def _pval_matrix(pvalues: np.ndarray, columns: list[str]):
    size = pvalues.shape[0]
    pvalues_str = np.zeros((size + 1, size + 1), dtype=np.dtypes.StringDType())
    for i in range(1, size + 1):
        for j in range(1, size + 1):
            if i > j:
                pvalues_str[i, j] = pvalue_to_asterisks(pvalues[i - 1, j - 1])
            elif i == j:
                pvalues_str[i, j] = "1.0"
            else:
                pvalues_str[i, j] = format(pvalues[i - 1, j - 1], ".5g")
    pvalues_str[0, 1:] = columns
    pvalues_str[1:, 0] = columns
    return pvalues_str



from __future__ import annotations

import numpy as np
from himena_stats._lazy_import import stats


def fit_norm(obs: np.ndarray) -> stats.rv_frozen:
    loc, scale = stats.norm.fit(obs)
    return stats.norm(loc=loc, scale=scale)


def fit_gamma(obs: np.ndarray) -> stats.rv_frozen:
    a, loc, scale = stats.gamma.fit(obs)
    return stats.gamma(a=a, scale=scale)


def fit_expon(obs: np.ndarray) -> stats.rv_frozen:
    loc, scale = stats.expon.fit(obs)
    return stats.expon(scale=scale)


def fit_uniform(obs: np.ndarray) -> stats.rv_frozen:
    # NOTE: scipy.stats uses [loc, loc + scale] to specify a uniform distribution.
    loc, scale = stats.uniform.fit(obs)
    return stats.uniform(loc=loc, scale=scale)


def fit_beta(obs: np.ndarray) -> stats.rv_frozen:
    a, b, loc, scale = stats.beta.fit(obs)
    return stats.beta(a=a, b=b)


def fit_cauchy(obs: np.ndarray) -> stats.rv_frozen:
    loc, scale = stats.cauchy.fit(obs)
    return stats.cauchy(loc=loc, scale=scale)


def fit_t(obs: np.ndarray) -> stats.rv_frozen:
    df, loc, scale = stats.t.fit(obs)
    return stats.t(df=df, loc=loc, scale=scale)


def fit_chi2(obs: np.ndarray) -> stats.rv_frozen:
    df, loc, scale = stats.chi2.fit(obs)
    return stats.chi2(df=df, loc=loc, scale=scale)


def fit_binom(obs: np.ndarray) -> stats.rv_frozen:
    mean = obs.mean()
    std = obs.std()
    p = 1 - std**2 / mean
    n = mean / p
    return stats.binom(n=n, p=p)


def fit_poisson(obs: np.ndarray) -> stats.rv_frozen:
    mu = obs.mean()
    return stats.poisson(mu=mu)


def fit_dist(obs, dist: stats.rv_frozen) -> stats.rv_frozen:
    dist_name: str = dist.dist.name
    if dist_name == "norm":
        return fit_norm(obs)
    elif dist_name == "gamma":
        return fit_gamma(obs)
    elif dist_name == "expon":
        return fit_expon(obs)
    elif dist_name == "uniform":
        return fit_uniform(obs)
    elif dist_name == "beta":
        return fit_beta(obs)
    elif dist_name == "cauchy":
        return fit_cauchy(obs)
    elif dist_name == "t":
        return fit_t(obs)
    elif dist_name == "chi2":
        return fit_chi2(obs)
    elif dist_name == "binom":
        return fit_binom(obs)
    elif dist_name == "poisson":
        return fit_poisson(obs)
    raise NotImplementedError



from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    from scipy import stats
    from numpy.typing import NDArray


def draw_pdf_or_pmf(
    dist: stats.rv_frozen,
    xlow: float,
    xhigh: float,
) -> tuple[NDArray[np.number], NDArray[np.number]]:
    if hasattr(dist, "pdf"):  # contiuous
        x0 = np.linspace(xlow, xhigh, 100)
        y0: np.ndarray = dist.pdf(x0)
        x = np.concatenate([[x0[0]], x0, [x0[-1]]])
        y = np.concatenate([[0.0], y0, [0.0]])
    elif hasattr(dist, "pmf"):  # discrete
        x0 = np.arange(int(xlow), int(xhigh) + 1)
        y0: np.ndarray = dist.pmf(x0)
        x = np.empty(4 * len(x0))
        x[0::4] = x[1::4] = x0 - 0.1
        x[2::4] = x[3::4] = x0 + 0.1
        y = np.zeros(4 * len(x0))
        y[1::4] = y0
        y[2::4] = y0
    else:
        raise TypeError(f"Type {type(dist)} not allowed.")
    return x, y


def infer_edges(dist: stats.rv_frozen) -> tuple[float, float]:
    if np.isfinite(dist.a):
        xlow = dist.a
    else:
        xlow = dist.ppf(0.001)
    if np.isfinite(dist.b):
        xhigh = dist.b
    else:
        xhigh = dist.ppf(0.999)
    return xlow, xhigh



from __future__ import annotations
from typing import Literal

from himena import Parametric, StandardType, WidgetDataModel
from himena.widgets import SubWindow
from himena.plugins import register_function, configure_gui
from himena.utils.table_selection import range_getter
from himena.qt.magicgui import SelectionEdit

from himena_stats._lazy_import import stats
from himena_stats.consts import MENUS_TEST
from himena_stats.test_tools._utils import (
    pvalue_to_asterisks,
    values_groups_to_xy,
    dropna,
)


@register_function(
    menus=MENUS_TEST,
    title="T-test ...",
    types=[StandardType.TABLE, StandardType.DATAFRAME, StandardType.EXCEL],
    command_id="himena-stats:test:t-test",
)
def t_test(win: SubWindow) -> Parametric:
    """Run a Student's or Welch's t-test on a table-like data."""

    @configure_gui(
        a={"widget_type": SelectionEdit, "getter": range_getter(win)},
        b={"widget_type": SelectionEdit, "getter": range_getter(win)},
        groups={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_t_test(
        a,
        b=None,
        groups=None,
        alternative: Literal["two-sided", "less", "greater"] = "two-sided",
        kind: Literal["Student", "Welch", "F"] = "Student",
        f_threshold: float = 0.05,
    ):
        model = win.to_model()
        x0, y0 = values_groups_to_xy(model, [a, b], groups)
        if kind == "F":
            f_result = stats.f_oneway(dropna(x0), dropna(y0))
            if f_result.pvalue < f_threshold:
                kind = "Student"
            else:
                kind = "Welch"
        t_result = stats.ttest_ind(
            dropna(x0), dropna(y0), equal_var=kind == "Student", alternative=alternative
        )
        return _ttest_result_to_model(
            t_result,
            title=f"T-test result of {model.title}",
            rows=[["kind", kind], ["comparison", f"{x0.name} vs {y0.name}"]],
        )

    return run_t_test


@register_function(
    menus=MENUS_TEST,
    title="Paired T-test ...",
    types=[StandardType.TABLE, StandardType.DATAFRAME, StandardType.EXCEL],
    command_id="himena-stats:test:paired-t-test",
)
def paired_t_test(win: SubWindow) -> Parametric:
    @configure_gui(
        a={"widget_type": SelectionEdit, "getter": range_getter(win)},
        b={"widget_type": SelectionEdit, "getter": range_getter(win)},
        groups={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_t_test(
        a,
        b=None,
        groups=None,
        alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    ):
        model = win.to_model()
        x0, y0 = values_groups_to_xy(model, [a, b], groups)
        t_result = stats.ttest_rel(dropna(x0), dropna(y0), alternative=alternative)
        return _ttest_result_to_model(
            t_result,
            title=f"Paired T-test result of {model.title}",
            rows=[["comparison", f"{x0.name} vs {y0.name}"]],
        )

    return run_t_test


@register_function(
    menus=MENUS_TEST,
    title="Wilcoxon Test ...",
    types=[StandardType.TABLE, StandardType.DATAFRAME, StandardType.EXCEL],
    command_id="himena-stats:test:wilcoxon-test",
)
def wilcoxon_test(win: SubWindow) -> Parametric:
    @configure_gui(
        a={"widget_type": SelectionEdit, "getter": range_getter(win)},
        b={"widget_type": SelectionEdit, "getter": range_getter(win)},
        groups={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_wilcoxon_test(
        a,
        b=None,
        groups=None,
        alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    ):
        model = win.to_model()
        x0, y0 = values_groups_to_xy(model, [a, b], groups)
        w_result = stats.wilcoxon(dropna(x0), dropna(y0), alternative=alternative)
        w_result_table = [
            ["p-value", format(w_result.pvalue, ".5g")],
            ["", pvalue_to_asterisks(w_result.pvalue)],
            ["statistic", format(w_result.statistic, ".5g")],
        ]
        return WidgetDataModel(
            value=w_result_table,
            type=StandardType.TABLE,
            title=f"Wilcoxon Test result of {model.title}",
        )

    return run_wilcoxon_test


@register_function(
    menus=MENUS_TEST,
    title="Mann-Whitney U Test ...",
    types=[StandardType.TABLE, StandardType.DATAFRAME, StandardType.EXCEL],
    command_id="himena-stats:test:mann-whitney-u-test",
)
def mann_whitney_u_test(win: SubWindow) -> Parametric:
    @configure_gui(
        a={"widget_type": SelectionEdit, "getter": range_getter(win)},
        b={"widget_type": SelectionEdit, "getter": range_getter(win)},
        groups={"widget_type": SelectionEdit, "getter": range_getter(win)},
    )
    def run_mann_whitney_u_test(
        a,
        b=None,
        groups=None,
        alternative: Literal["two-sided", "less", "greater"] = "two-sided",
    ):
        model = win.to_model()
        x0, y0 = values_groups_to_xy(model, [a, b], groups)
        u_result = stats.mannwhitneyu(dropna(x0), dropna(y0), alternative=alternative)
        u_result_table = [
            ["p-value", format(u_result.pvalue, ".5g")],
            ["", pvalue_to_asterisks(u_result.pvalue)],
            ["U-statistic", format(u_result.statistic, ".5g")],
        ]
        return WidgetDataModel(
            value=u_result_table,
            type=StandardType.TABLE,
            title=f"Mann-Whitney U Test result of {model.title}",
        )

    return run_mann_whitney_u_test


def _ttest_result_to_model(t_result, title: str, rows: list[list[str]] = ()):
    t_result_table = [
        ["p-value", format(t_result.pvalue, ".5g")],
        ["", pvalue_to_asterisks(t_result.pvalue)],
        ["t-statistic", format(t_result.statistic, ".5g")],
        ["degrees of freedom", int(t_result.df)],
    ] + rows
    return WidgetDataModel(
        value=t_result_table,
        type=StandardType.TABLE,
        title=title,
    )



import pytest
from himena import MainWindow, StandardType

table_data_sig = [
    ["a", "b"],
    [1.9, 4.4],
    [2.1, 5.3],
    [1.6, 4.6],
    [2.1, 4.1],
    [1.8, 5.9],
    [1.7, 5.3],
    ["", 5.1],
]


table_data_nonsig = [
    ["a", "b"],
    [4.9, 4.8],
    [5.1, 5.3],
    [4.6, 4.6],
    [5.1, 5.0],
    [5.8, 5.9],
    [5.7, 5.6],
    ["", 5.1],
]

def unpivot(val):
    out = []
    for r0, _ in val[1:]:
        if r0 == "":
            continue
        out.append(["a", r0])
    for _, r1 in val[1:]:
        out.append(["b", r1])
    return out

a_val = ((0, 100), (0, 1))
b_val = ((0, 100), (1, 2))

@pytest.mark.parametrize(
    "command",
    ["t-test", "mann-whitney-u-test"]
)
def test_a_vs_b(himena_ui: MainWindow, command: str):
    win_sig = himena_ui.add_object(table_data_sig, type=StandardType.TABLE)
    win_sig_unpivot = himena_ui.add_object(
        unpivot(table_data_sig), type=StandardType.TABLE
    )
    win_nonsig = himena_ui.add_object(table_data_nonsig, type=StandardType.TABLE)
    win_nonsig_unpivot = himena_ui.add_object(
        unpivot(table_data_nonsig), type=StandardType.TABLE
    )
    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_sig,
        with_params={"a": a_val, "b": b_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) < 0.1

    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_nonsig,
        with_params={"a": a_val, "b": b_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) > 0.3

    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_sig_unpivot,
        with_params={"a": b_val, "groups": a_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) < 0.1

    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_nonsig_unpivot,
        with_params={"a": b_val, "groups": a_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) > 0.3

@pytest.mark.parametrize(
    "command",
    ["paired-t-test", "wilcoxon-test"],
)
def test_a_vs_b_same_shape(himena_ui: MainWindow, command: str):
    win_sig = himena_ui.add_object(table_data_sig[:-1], type=StandardType.TABLE)
    win_sig_unpivot = himena_ui.add_object(
        unpivot(table_data_sig[:-1]), type=StandardType.TABLE
    )
    win_nonsig = himena_ui.add_object(table_data_nonsig[:-1], type=StandardType.TABLE)
    win_nonsig_unpivot = himena_ui.add_object(
        unpivot(table_data_nonsig[:-1]), type=StandardType.TABLE
    )
    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_sig,
        with_params={"a": a_val, "b": b_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) < 0.1

    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_nonsig,
        with_params={"a": a_val, "b": b_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) > 0.3

    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_sig_unpivot,
        with_params={"a": b_val, "groups": a_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) < 0.1

    himena_ui.exec_action(
        f"himena-stats:test:{command}",
        window_context=win_nonsig_unpivot,
        with_params={"a": b_val, "groups": a_val}
    )
    assert himena_ui.current_model.value[0, 0] == "p-value"
    assert float(himena_ui.current_model.value[0, 1]) > 0.3



from typing import TypeVar
import numpy as np
from numpy.typing import NDArray
from magicgui.widgets.bases import ValueWidget
from himena import Parametric, StandardType, WidgetDataModel
from himena.widgets import SubWindow
from himena.plugins import register_function, configure_gui
from himena.data_wrappers import wrap_dataframe
from himena.standards import plotting as hplt
from himena.standards.model_meta import TableMeta
from himena_stats._lazy_import import stats
from himena_stats.consts import MENUS_DIST
from himena_stats.distributions._utils import draw_pdf_or_pmf, infer_edges
from himena_stats.distributions._fit import fit_dist

OBS_TYPES = [StandardType.TABLE, StandardType.ARRAY, StandardType.DATAFRAME]


@register_function(
    menus=MENUS_DIST,
    title="Fit Distribution ...",
    types=StandardType.DISTRIBUTION,
    command_id="himna-stats:dist-convert:fit-mle",
)
def fit_mle(model: WidgetDataModel) -> Parametric:
    """Fit distribution model to observations by maximum likelihood estimation."""

    @configure_gui(
        obs={"types": OBS_TYPES, "label": "observations"},
        obs_range={"bind": _get_range},
    )
    def run_fit(
        obs: WidgetDataModel,
        obs_range: tuple[tuple[int, int], tuple[int, int]],
    ) -> WidgetDataModel:
        """
        Parameters
        ----------
        obs : WidgetDataModel, optional
            Observations to which model will be fit. This value can be a table or an
            array, optionally with a selection area that specifies the range.
        param_as_guess : bool, default False
            If checked, the current parameter of the distribution will be used as the
            initial guess. Otherwise, only the distribution model will be considered.
        """
        dist: "stats.rv_frozen" = model.value
        dtype = np.float64 if hasattr(dist, "pdf") else np.int64
        arr = _norm_obs(obs, obs_range, np.dtype(dtype))
        dist_fitted = fit_dist(arr, dist)
        return WidgetDataModel(
            value=dist_fitted,
            type=StandardType.DISTRIBUTION,
            title=f"{model.title} fitted",
        )

    return run_fit


@register_function(
    menus=MENUS_DIST,
    title="Plot Distribution ...",
    types=StandardType.DISTRIBUTION,
    command_id="himena-stats:dist-plot:plot",
)
def plot_dist(win: SubWindow) -> Parametric:
    """Plot distribution with observations."""

    @configure_gui(
        obs={"types": OBS_TYPES, "label": "observations"},
        obs_range={"bind": _get_range},
    )
    def run_plot(obs: WidgetDataModel | None, obs_range) -> WidgetDataModel:
        """Plot distribution, optional with the observations.

        Parameters
        ----------
        obs : WidgetDataModel, optional
            If given, this observation data will also be plotted as a histogram. This
            value can be a table or an array, optionally with a selection area that
            specifies the range to plot.
        """
        model = win.to_model()
        dist: "stats.rv_frozen" = model.value
        xlow, xhigh = infer_edges(dist)
        is_continuous = hasattr(dist, "pdf")
        fig = hplt.figure()
        if obs is not None:
            dtype = np.float64 if is_continuous else np.int64
            arr = _norm_obs(obs, obs_range, np.dtype(dtype))
            if is_continuous:
                fig.hist(
                    arr,
                    bins=min(int(np.sqrt(arr.size)), 64),
                    stat="density",
                    color="skyblue",
                    name="observations",
                )
            else:
                values, counts = np.unique_counts(arr)
                density = counts / arr.size
                fig.bar(values, density, color="skyblue", name="observations")
            xlow = min(xlow, arr.min())
            xhigh = max(xhigh, arr.max())

        x, y = draw_pdf_or_pmf(dist, xlow, xhigh)
        fig.plot(x, y, width=2.5, color="red", name="model")
        return WidgetDataModel(
            value=fig,
            type=StandardType.PLOT,
            title=f"Plot of {model.title}",
        )

    return run_plot


@register_function(
    menus=MENUS_DIST,
    title="Random Sampling ...",
    types=StandardType.DISTRIBUTION,
    command_id="himena-stats:dist:sample",
)
def sample_dist(win: SubWindow) -> Parametric:
    """Random sampling from a distribution."""
    random_state_default = np.random.randint(0, 10000)

    @configure_gui(random_state={"value": random_state_default})
    def run_sample(sample_size: list[int] = (100,), random_state: int | None = None):
        """Sample from the distribution.

        Parameters
        ----------
        sample_size : list of int
            The size of the sample. Can be a verctor for n-dimentional sampling.
        """
        model = win.to_model()
        dist: "stats.rv_frozen" = model.value
        samples = dist.rvs(size=sample_size, random_state=random_state)
        return WidgetDataModel(
            value=samples,
            type=StandardType.ARRAY,
            title=f"Samples from {model.title}",
        )

    return run_sample


@register_function(
    menus=MENUS_DIST,
    title="Show statistics",
    types=StandardType.DISTRIBUTION,
    command_id="himena-stats:dist:stats",
)
def show_stats(model: WidgetDataModel) -> WidgetDataModel:
    """Show statistics of the distribution."""
    dist: "stats.rv_frozen" = model.value
    value = [
        ["min", dist.a],
        ["mean", dist.mean()],
        ["median", dist.median()],
        ["max", dist.b],
        ["std", dist.std()],
        ["entropy", dist.entropy()],
    ]
    return WidgetDataModel(
        value=value,
        type=StandardType.TABLE,
        title=f"Statistics of {model.title}",
    )


def _get_range(widget: ValueWidget):
    model: WidgetDataModel = widget.parent["obs"].value
    if not isinstance(meta := model.metadata, TableMeta):
        return None
    if len(meta.selections) != 1:
        raise ValueError(f"Data {model.title} must contain single selection.")
    return meta.selections[0]


@register_function(
    menus=MENUS_DIST,
    title="Plot CDF",
    types=StandardType.DISTRIBUTION,
    command_id="himena-stats:dist-plot:cdf",
)
def plot_cdf(model: WidgetDataModel) -> WidgetDataModel:
    """Plot the cumulative distribution function of the model."""
    dist: "stats.rv_frozen" = model.value
    xlow, xhigh = infer_edges(dist)
    x = np.linspace(xlow, xhigh, 256)
    y = dist.cdf(x)
    fig = hplt.figure()
    fig.plot(x, y, color="red", name=model.title)
    return WidgetDataModel(
        value=fig, type=StandardType.PLOT, title=f"CDF of {model.title}"
    )


@register_function(
    menus=MENUS_DIST,
    title="Plot Survival Function",
    types=StandardType.DISTRIBUTION,
    command_id="himena-stats:dist-plot:sf",
)
def plot_sf(model: WidgetDataModel) -> WidgetDataModel:
    """Plot the survival function of the model."""
    dist: "stats.rv_frozen" = model.value
    xlow, xhigh = infer_edges(dist)
    x = np.linspace(xlow, xhigh, 256)
    y = dist.sf(x)
    fig = hplt.figure()
    fig.plot(x, y, color="red", name=model.title)
    return WidgetDataModel(
        value=fig, type=StandardType.PLOT, title=f"Survival function of {model.title}"
    )


_D = TypeVar("_D", bound=np.generic)


def _norm_obs(obs: WidgetDataModel, obs_range, dtype: np.dtype[_D]) -> NDArray[_D]:
    if obs_range is None:
        obs_slice = slice(None)
    else:
        rinds, cinds = obs_range
        obs_slice = slice(*rinds), slice(*cinds)
    if obs.is_subtype_of(StandardType.TABLE):
        arr = obs.value[obs_slice].astype(dtype)
    elif obs.is_subtype_of(StandardType.DATAFRAME):
        rsl, csl = obs_slice
        if csl.start - csl.stop != 1:
            raise ValueError("Only single-column selection is allowed")
        df = wrap_dataframe(obs.value)
        arr = df[rsl, csl.start].astype(dtype)
    elif obs.is_subtype_of(StandardType.ARRAY):
        arr = obs.value[obs_slice].astype(dtype)
    else:
        raise NotImplementedError
    return arr
